/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation version
 * 2.1 of the License.
 *                                                                                  
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *                                                                                  
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * END COPYRIGHT BLOCK **/
#include <stdio.h>
#ifdef XP_UNIX
#include <libgen.h>
#endif
#include "ldap.h"
#include "ldap_ssl.h"
#include "libadmsslutil/psetcssl.h"
#include "libadmsslutil/admsslutil.h"
#include "cert.h"
#include "secmod.h"
#include "ssl.h"
#include "nss.h"
#include "pkcs11.h"
#include "secpwd_pvt.h"
#include "prprf.h"
#include "sslproto.h"

#include "errno.h"
#include "secport.h"
#include "libadmsslutil/certmgt.h"
#include "version.h"

#define ADMSSL_BUF_LEN 1024

#define IDEDIT 1000
#define IDD_PASSWORD 108
#define IDTOKEN 1004
                   
#ifndef FILE_PATHSEP
#define FILE_PATHSEP '/'
#endif

#define SECURITY_ATTR "configuration.nsServerSecurity"

#ifdef XP_UNIX
/*
 * Version information for the 'ident' and 'what commands
 *
 * NOTE: the first component of the concatenated rcsid string
 * must not end in a '$' to prevent rcs keyword substitution.
 */
static char rcsid[] = "$Header: AdminUtil " ADMSDK_PRODUCT  
        "  " ADMSDK_BUILDNUM " $";
static char sccsid[] = "@(#)AdminUtil " ADMSDK_PRODUCT  
		"  " ADMSDK_BUILDNUM;
#endif /* XP_UNIX */

void servssl_error(char *fmt, ...);

char custom_file[BIG_LINE] = "";
char retried_token[BIG_LINE] = "";

PR_IMPLEMENT(int)
admldapBuildInfoSSL(AdmldapInfo info, int *errorcode)
{
  LDAP  *ld;
  int   ldapError;
  char *passwd = NULL;
  char *host = NULL;

  *errorcode = ADMUTIL_OP_OK;

  host = admldapGetHost(info);
  if (admldapGetSecurity(info))
    ld = ldapssl_init(host, admldapGetPort(info), 1);
  else
    ld = ldap_init(host, admldapGetPort(info));
  PL_strfree(host);

  if (!ld) {
    *errorcode = ADMUTIL_LDAP_ERR;
    return 0;
  }

  passwd = admldapGetSIEPWD(info);
  if (passwd) {
      char *dn = admldapGetSIEDN(info);
      ldapError = ldap_simple_bind_s(ld, dn, passwd);
      PL_strfree(dn);
      PL_strfree(passwd);
  } else {
      /* no password means just punt rather than do anon bind */
      /* this mimics the same logic in admldapBuildInfoCbk() */
      *errorcode = ADMUTIL_LDAP_ERR;
      return 1; /* have to return true here to mimic admldapBuildInfoCbk() */
  }
  /* authenticate to LDAP server*/
  if (ldapError != LDAP_SUCCESS) {
#ifdef LDAP_DEBUG
    ldap_perror(ld, "ldap_simple_bind_s");
#endif
    switch (ldapError) {
    case LDAP_INAPPROPRIATE_AUTH:
    case LDAP_INVALID_CREDENTIALS:
    case LDAP_INSUFFICIENT_ACCESS:
      /* authenticate failed: Should not continue */
      ldap_unbind(ld);
      *errorcode = ADMUTIL_LDAP_ERR;
      return 0;
    case LDAP_NO_SUCH_OBJECT:
    case LDAP_ALIAS_PROBLEM:
    case LDAP_INVALID_DN_SYNTAX:
      /* Not a good user DN */
      ldap_unbind(ld);
      *errorcode = ADMUTIL_LDAP_ERR;
      return 0;
      break;
    default:
      ldap_unbind(ld);
      *errorcode = ADMUTIL_LDAP_ERR;
      return 0;
    }
  }

  admldapSetLDAPHndl(info, ld);
  return 1;
}

/*
 * Given filepath e.g. /opt/servers/alias/slapd-dilly-cert7.db
 * return slapd-dilly- as the prefix fo cert file;
 * or filepath is e.g. /opt/servers/alias/slapd-dilly-key3.db
 * return slapd-dilly- as the prefix for key file.
*/
static void
getPrefix(char *filepath, char *basename, char *prefix, int maxprefixlen)
{
    char *filename, *base;
    int prefixLen;

    if (NULL == prefix || 0 == maxprefixlen) {
        return;
    }
    prefix[0] = '\0';

    if((filename = strrchr(filepath, FILE_PATHSEP)) == NULL) {
        filename=filepath;
    }
    else {
        filename++; /* Skip '/' */
    }

    /* basename is either "-cert" or "-key" - look for last occurance */
    base = PL_strrstr(filename, basename);
    if (base == NULL) {
        return;
    }

    /* Include '-' into prefix */
    prefixLen = base-filename+1;
    if (prefixLen > maxprefixlen) {
        return;
    }
    memcpy(prefix, filename, prefixLen);
    prefix[prefixLen]=0;
}

static int initNSS(char *certdbFile, char *keydbFile)
{

  char certdbPrefix[256], keydbPrefix[256];
  char secdbDir[PATH_MAX];
  char admconfDir[PATH_MAX];
  PRUint32 flags = 0;
  char *db_name;

  getPrefix(certdbFile, "-cert", certdbPrefix, 256);
  getPrefix(keydbFile, "-key", keydbPrefix, 256);

  if(('\0' == certdbPrefix) || ('\0' == keydbPrefix)) return -1;

  PR_snprintf(secdbDir, sizeof(secdbDir), "%s%calias",
              getenv("NETSITE_ROOT"), FILE_PATHSEP);
  PR_snprintf(admconfDir, sizeof(admconfDir), "%s%cadmin-serv%cconfig",
              getenv("NETSITE_ROOT"), FILE_PATHSEP,FILE_PATHSEP);

  /* PKSC11 module must be configured before NSS is initialized */
  db_name = PL_strdup("internal (software)              ");
  if (!db_name) {
      return -1;
  }

  PK11_ConfigurePKCS11(NULL,NULL,NULL,db_name,NULL, NULL,NULL,NULL,
                       /*minPwdLen=*/8, /*pwdRequired=*/1);
 
  /* init NSS */ 
  if (NSS_Initialize(secdbDir, certdbPrefix, keydbPrefix, "secmod.db", flags)) {
      return -1;
  }

  /* set password function */
  PR_snprintf(custom_file, sizeof(custom_file), "%s%ccustom.conf",
              admconfDir, FILE_PATHSEP);
  PK11_SetPasswordFunc(getPassword_cb);

  /* enable all default ciphers */
  if (SSL_OptionSetDefault(SSL_ENABLE_SSL2, PR_FALSE)
      || SSL_OptionSetDefault(SSL_ENABLE_SSL3, PR_TRUE)) {
      return -1;
  }

  /* set export policy */
  if(SSLPLCY_Install() != PR_SUCCESS) {
    return -1;
  }

  return 0;
}


PR_IMPLEMENT(int)
ADMSSL_InitSimple2(char* serverRoot)
{
  AdmldapInfo   admLdapInfo=NULL;
  char configDir[256];
  int  error;

  PR_snprintf(configDir, sizeof(configDir), "%s%cadmin-serv%cconfig",
              serverRoot, FILE_PATHSEP, FILE_PATHSEP);
  admLdapInfo = admldapBuildInfo(configDir, &error);
  if (!admLdapInfo) return -1;

  if (!admldapGetSecurity(admLdapInfo)) return 0;

  error = ADMSSL_Init(admLdapInfo, configDir);
  destroyAdmldap(admLdapInfo);
  return error;
}

PR_IMPLEMENT(int)
ADMSSL_InitSimple()
{
  char *serverRoot = getenv("NETSITE_ROOT");

  return ADMSSL_InitSimple2(serverRoot);
}
  

/* set minimum SSL stuff for LDAP/SSL to work */
PR_IMPLEMENT(int)
ADMSSL_Init(AdmldapInfo info, char *admroot)
{
  char *certdbFile, *keydbFile;
  int   secure;
  char *dirURL, *bindDN, *bindPwd, *dirInfoRef; 
  int errCode;

  secure = admldapGetSecurity(info);

  if (!secure) {
    admldapGetLocalUserDirectory(info,
                                 &dirURL, 
                                 &bindDN, 
                                 &bindPwd, 
                                 &dirInfoRef, 
                                 &errCode);
    if (errCode || !dirURL) return 0;
    if (!strstr(dirURL, "ldaps://"))
        return 0;
  }

  certdbFile = admldapGetCertDBFile(info);
  keydbFile = admldapGetKeyDBFile(info);

  if((!certdbFile) || (!keydbFile) || (!admroot)) return -1;

  errCode = initNSS(certdbFile, keydbFile);
  PR_Free(certdbFile);
  PR_Free(keydbFile);

  return errCode;
}

void servssl_error(char *fmt, ...)
{
    va_list args;
    char errstr[ADMSSL_BUF_LEN];

    va_start(args, fmt);
    PR_vsnprintf(errstr, sizeof(errstr), fmt, args);
#ifdef XP_UNIX
    fprintf(stderr, "%s\n", errstr);
#else  /* XP_WIN32 */
    MessageBox(NULL, errstr,
               "SSL_INITIALIZATION_ERROR", MB_APPLMODAL|MB_ICONEXCLAMATION|MB_OK);
#endif
    exit(1);
}

void servssl_warn(char *fmt, ...)
{
    va_list args;
    char warnstr[ADMSSL_BUF_LEN];

    va_start(args, fmt);
    PR_vsnprintf(warnstr, sizeof(warnstr), fmt, args);
#ifdef XP_UNIX
    fprintf(stderr, "%s\n", warnstr);
#else  /* XP_WIN32 */
    MessageBox(NULL, warnstr,
               "SSL_INITIALIZATION_WARNING", MB_APPLMODAL|MB_ICONEXCLAMATION|MB_OK);
#endif
}

/* ----------------------- Multiple cipher support ------------------------ */

#ifdef NET_SSL

typedef struct {
    char *name;
    int num;
} cipherstruct;

static cipherstruct _conf_ciphers[] = {
    {"rc4", SSL_EN_RC4_128_WITH_MD5},
    {"rc4export", SSL_EN_RC4_128_EXPORT40_WITH_MD5},
    {"rc2", SSL_EN_RC2_128_CBC_WITH_MD5},
    {"rc2export", SSL_EN_RC2_128_CBC_EXPORT40_WITH_MD5},
    {"idea", SSL_EN_IDEA_128_CBC_WITH_MD5},
    {"des", SSL_EN_DES_64_CBC_WITH_MD5},
    {"desede3", SSL_EN_DES_192_EDE3_CBC_WITH_MD5},
    {"rsa_rc4_128_md5", SSL_RSA_WITH_RC4_128_MD5},
    {"rsa_3des_sha", SSL_RSA_WITH_3DES_EDE_CBC_SHA},
    {"rsa_des_sha", SSL_RSA_WITH_DES_CBC_SHA},
    {"rsa_rc4_40_md5", SSL_RSA_EXPORT_WITH_RC4_40_MD5},
    {"rsa_rc2_40_md5", SSL_RSA_EXPORT_WITH_RC2_CBC_40_MD5},
    {"tls_rsa_export1024_with_des_cbc_sha", TLS_RSA_EXPORT1024_WITH_DES_CBC_SHA},
    {"tls_rsa_export1024_with_rc4_56_sha", TLS_RSA_EXPORT1024_WITH_RC4_56_SHA},
    {"rsa_null_md5", SSL_RSA_WITH_NULL_MD5},
    {"fortezza", SSL_FORTEZZA_DMS_WITH_FORTEZZA_CBC_SHA},
    {"fortezza_rc4_128_sha", SSL_FORTEZZA_DMS_WITH_RC4_128_SHA},
    {"fortezza_null", SSL_FORTEZZA_DMS_WITH_NULL_SHA},
    {"rsa_fips_des_sha", SSL_RSA_FIPS_WITH_DES_CBC_SHA},
    {"rsa_fips_3des_sha", SSL_RSA_FIPS_WITH_3DES_EDE_CBC_SHA},
    {NULL, 0}
};

void _conf_setallciphers(int active)
{
    register int x;

    /* MLM - change: Because null_md5 is NOT encrypted at all, force
     *       them to activate it by name. */
    for(x = 0; _conf_ciphers[x].name; x++)  {
        if(active && !strcmp(_conf_ciphers[x].name, "rsa_null_md5"))  {
            continue;
        }
#ifdef NET_SSL
        SSL_CipherPrefSetDefault(_conf_ciphers[x].num, active ? PR_TRUE : PR_FALSE);
#endif
    }
}

char *_conf_setciphers(char *ciphers)
{
    char *t, err[ADMSSL_BUF_LEN];
    int x, active;

    /* Default is to activate all of them */
    if(!ciphers) {
        _conf_setallciphers(1);
        return NULL;
    }
    _conf_setallciphers(0);

    t = ciphers;
    while(t) {
        while((*ciphers) && (isspace(*ciphers))) ++ciphers;

        switch(*ciphers++) {
          case '+':
            active = 1; break;
          case '-':
            active = 0; break;
          default:
            return(PL_strdup("invalid ciphers: format is +cipher1,-cipher2..."));
        }
        if( (t = strchr(ciphers, ',')) )
            *t++ = '\0';

        if(!strcmp(ciphers, "all"))
            _conf_setallciphers(active);
        else {
            for(x = 0; _conf_ciphers[x].name; x++) {
                if(!strcmp(ciphers, _conf_ciphers[x].name)) {
                  SSL_CipherPrefSetDefault(_conf_ciphers[x].num, active ? PR_TRUE : PR_FALSE);
                    break;
                }
            }
            if(!_conf_ciphers[x].name) {
                PR_snprintf(err, sizeof(err), "unknown cipher %s", ciphers);
                return PL_strdup(err);
            }
        }
        if(t)
            ciphers = t;
    }
    return NULL;
}


/*
 * SSLPLCY_Install
 */
PRStatus
SSLPLCY_Install(void)
{
  SECStatus s;
  /* There is no more export builds */
  s = NSS_SetDomesticPolicy();
  return (s==SECSuccess)?PR_SUCCESS:PR_FAILURE;
}


#endif /* NET_SSL */


/*
 *  In HCL 1.5 final this is called by each socket created.
 *  Pass the socket in. - adam
 */
PR_IMPLEMENT(void)
servssl_init(PsetHndl pset, PRFileDesc *fd, char *admroot)
{
    SECKEYPrivateKey  *key = NULL;
    CERTCertificate   *cert = NULL;
    int               rv; 

    int errorCode;
    AttrNameList family_list;

    char *keyfn = NULL;
    char *certfn = NULL;
    char *val = NULL;
    char *val2 = NULL;
    int stimeout;
    long ssl3timeout;
#define CIPHER_STRLEN 1024
    char cipher_string[CIPHER_STRLEN];
    int cipher_len = CIPHER_STRLEN;

    if (!pset) {
      servssl_error("Security Initialization: Failed to retrieve SSL cipher family information\n");
      return;
    }

    /* Get general pset information */    
    
    keyfn =  psetGetAttrSingleValue(pset, 
                                    "configuration.encryption.nsKeyfile", 
                                    &errorCode);

    certfn =  psetGetAttrSingleValue(pset, 
                                     "configuration.encryption.nsCertfile", 
                                     &errorCode);

    val = psetGetAttrSingleValue(pset, 
                                 "configuration.encryption.nsSslSessionTimeout", 
                                 &errorCode);

    val2 = psetGetAttrSingleValue(pset, 
                                  "configuration.encryption.nsSsl3SessionTimeout", 
                                  &errorCode);

    if((!keyfn) || (!certfn) || (!val) || (!val2)) {
      servssl_error("Security Initialization: Failed to retrieve SSL cipher family information (error code = %d)\n", errorCode);
      return;
    }      

    stimeout = atoi(val);
    ssl3timeout = atol(val2);

    /* Initialize the cert data base  */
    if (initNSS(certfn, keyfn)) {
      servssl_error("Security Initialization: Failed to initialize NSS\n");
        return;
    }
    
    /* Set SSL cipher preferences */
    cipher_string[0] = '\0';
    val = NULL;
    val = psetGetAttrSingleValue(pset, 
                                 "configuration.encryption.nsSsl2Ciphers", 
                                 &errorCode);    
    if(val && strcmp(val, "blank")) {
      PL_strncat(cipher_string, val, cipher_len);
      cipher_len -= PL_strlen(val);
    }

    val = NULL;
    val = psetGetAttrSingleValue(pset, 
                                 "configuration.encryption.nsSsl3Ciphers", 
                                 &errorCode);    
    if(val && strcmp(val, "blank")) {
      if(cipher_string[0] != '\0') {
        PL_strncat(cipher_string, ",", cipher_len);
        cipher_len -= 1;
      }
      PL_strncat(cipher_string, val, cipher_len);
      cipher_len -= PL_strlen(val);
    }

    if(!strcmp(cipher_string, ""))
      _conf_setciphers(NULL);
    else {
      if(_conf_setciphers(cipher_string))
        servssl_warn("Security Initialization Warning: Failed to set SSL cipher preference information!");
    }


    /* 
     * Now, get the complete list of cipher families. Each family
     * has a token name and personality name which we'll use to find
     * appropriate keys and certs, and call SSL_ConfigSecureServer
     * with.
     */
    
    if((family_list = psetGetChildren(pset, "configuration.Encryption", &errorCode))) {
      char **family;
      char cert_name[ADMSSL_BUF_LEN];
      char family_attribute[ADMSSL_BUF_LEN];
      char *token;
      char *personality;
      PK11SlotInfo *current_slot;

      for (family = family_list; *family; family++) {

        token = NULL;
        personality = NULL;
        val = NULL;

        PR_snprintf(family_attribute, sizeof(family_attribute),
                    "%s.nsSslActivation", *family);        
        val =  psetGetAttrSingleValue(pset, 
                                      family_attribute, 
                                      &errorCode);

        if((!val) || (!strcmp(val, "off")))
          /* this family was turned off, goto next */
          continue;
        
        PR_snprintf(family_attribute, sizeof(family_attribute),
                    "%s.nsSslToken", *family);        
        token =  psetGetAttrSingleValue(pset, 
                                        family_attribute, 
                                        &errorCode);
        PR_snprintf(family_attribute, sizeof(family_attribute),
                    "%s.nsSslPersonalityssl", *family);        
        personality =  psetGetAttrSingleValue(pset, 
                                              family_attribute, 
                                              &errorCode);

        if((token) && (personality)) {
          if(strstr(token, "internal")) {
            PL_strncpyz(cert_name, personality, sizeof(cert_name));
            current_slot = PK11_GetInternalKeySlot();
          }
          else {
            /* external PKCS #11 token - attach token name */
            PR_snprintf(cert_name, sizeof(cert_name), "%s:%s",
                        token, personality);
            current_slot = PK11_FindSlotByName(token);
          }
        }
        else {
          servssl_error("Security Initialization: Failed to get cipher family information");
        }

        /* Step Three.6 - If in FIPS mode, authenticate to the token before doing anything else */
        if(PK11_IsFIPS()) {
          PK11_Authenticate(current_slot, PR_FALSE, NULL);
        }

        PK11_SetSlotPWValues(current_slot, 0, 0);

        /* Step Four -- Locate the server certificate */
        cert = PK11_FindCertFromNickname(cert_name, NULL);
        
        if (cert == NULL) {
          servssl_error("Security Initialization: Can't find certificate (%s)", personality);
        }

        /* Step Four.5 -- check that the cert is valid */
        if(CERT_VerifyCertNow(/*certdb=*/NULL, cert, PR_FALSE, certUsageSSLServer, NULL) != SECSuccess) {
          servssl_warn("Security Initialization Warning: For certificate (%s) - %s", 
                       personality, SSL_Strerror(PR_GetError()));
        }

        /* Step Five -- Get the private key from cert  */
        key = PK11_FindKeyByAnyCert(cert, NULL);
        
        if (key == NULL) {
          servssl_error("Security Initialization: Unable to retrieve private key (%d)", PR_GetError());
        }        

        /* Step Six  -- Configure Secure Server Mode  */
        if(fd) {
          if(PK11_FortezzaHasKEA(cert) == PR_TRUE)
            rv = SSL_ConfigSecureServer(fd, cert, key, kt_fortezza);
          else
            rv = SSL_ConfigSecureServer(fd, cert, key, kt_rsa);
        }        
        if (rv) {
          servssl_error("Security Initialization: Server key/certificate is bad (%d)", 
                       PR_GetError());
        }
      }
    }

    /* Step Seven -- Configure Server Session ID Cache  */
    val = (char *)PR_Malloc(L_tmpnam * sizeof(char));
#ifdef XP_WIN32
    if(GetTempPath(L_tmpnam, val) == 0)
       servssl_error("Security Initialization: Config of server nonce cache failed, "
                     "cannot find a valid temporary directory! (%d)", GetLastError());
    rv = SSL_ConfigServerSessionIDCache(0, stimeout, ssl3timeout, val);
#else
    rv = SSL_ConfigServerSessionIDCache(0, stimeout, ssl3timeout, dirname(tmpnam(val)));
#endif
    PR_Free(val);
    if (rv) {
      if (PR_GetError() == ENOSPC) {
        servssl_error("Security Initialization: Config of server nonce cache failed, "
                      "out of disk space! Make more room in /tmp "
                      "and try again.");
        } 
        else {
          servssl_error("Security Initialization: Config of server nonce cache failed (%d)",
                        PR_GetError());
        }
    }

    /* Question- Any house cleaning here???? key, keydb, cert and certdb */
}


/*
 * Modify "security" in adm.conf and DS
 */
void set_security(PsetHndl pset, 
                  char *sroot,
                  char *security, 
                  char *cert_file, 
                  char *key_file)
{
  FILE *f;
  int i, modified_security=0, modified_cert=0, modified_key=0;
  static char filename[BIG_LINE];
  static char inbuf[BIG_LINE];
  static char buf[BIG_LINE];
  int linecnt=0;
  char *col;
  char *lines[50];
  int rv;
  
  /* set security attribute in DS */
  if((pset) && (security) && (*security != '\0')) {
    if(psetCheckAttribute(pset, SECURITY_ATTR) == PSET_OP_FAIL) {
      /* create entry */
      rv = psetAddSingleValueAttribute(pset, SECURITY_ATTR, security);
      if(rv != PSET_OP_OK)
        servssl_error("PSET attribute creation failed!");      
    }
    else {
      /* modify entry */
      rv = psetSetSingleValueAttr(pset, SECURITY_ATTR, security);
      if(rv != PSET_OP_OK)
        servssl_error("PSET attribute modification failed!");
    }
  }

  /* add/edit security field in adm.conf */
  for(i=0; i<50; i++)
    lines[i] = NULL;

  PR_snprintf(filename, sizeof(filename),
              "%s/admin-serv/config/adm.conf", sroot);
  
  f = fopen(filename, "r");
  if (f==NULL) {
    fclose(f);
    servssl_error("Can not open adm.conf for reading");
  }
  
  while(fgets(inbuf, sizeof(inbuf), f) != NULL) {
    if ((strstr(inbuf,"security:") == inbuf) &&
        (security && *security != '\0')) { 
      /* Line starts with "security" */
      col = strrchr(inbuf,':');
      if (col == NULL) {
        servssl_error("Bad format for adminurl in adm.conf");
      }
      *col=0;
      PR_snprintf(buf, sizeof(buf), "%s: %s\n", inbuf, security);
      lines[linecnt++] = PL_strdup(buf);
      modified_security=1;
    }
    else if ((strstr(inbuf,"certDBFile:") == inbuf) &&
             (cert_file && *cert_file != '\0')) { 
      /* Line starts with "certDBFile" */
      col = strchr(inbuf,':');
      if (col == NULL) {
        servssl_error("Bad format for adminurl in adm.conf");
      }
      *col=0;
      PR_snprintf(buf, sizeof(buf), "%s: %s%c%s\n", 
                  inbuf, sroot, FILE_PATHSEP, cert_file);
      lines[linecnt++] = PL_strdup(buf);
      modified_cert=1;
    }
    else if ((strstr(inbuf,"keyDBFile:") == inbuf) &&
             (key_file && *key_file != '\0')) { 
      /* Line starts with "keyDBFile" */
      col = strchr(inbuf,':');
      if (col == NULL) {
        servssl_error("Bad format for adminurl in adm.conf");
      }
      *col=0;
      PR_snprintf(buf, sizeof(buf), "%s: %s%c%s\n", 
                  inbuf, sroot, FILE_PATHSEP, key_file);
      lines[linecnt++] = PL_strdup(buf);
      modified_key=1;
    }
    else {
      lines[linecnt++] = PL_strdup(inbuf);
    }
  }        
  fclose(f);
  
  if ((!modified_security) &&
      (security && *security != '\0')) {
    /* security not found - put it in */
    PR_snprintf(buf, sizeof(buf), "security: %s\n", security);
    lines[linecnt++] = PL_strdup(buf);
  }
  if ((!modified_cert) &&
      (cert_file && *cert_file != '\0')) {
    /* certDBFile not found - put it in */
    PR_snprintf(buf, sizeof(buf), "certDBFile: %s%c%s\n", 
                sroot, FILE_PATHSEP, cert_file);
    lines[linecnt++] = PL_strdup(buf);
  }
  if ((!modified_key) &&
      (key_file && *key_file != '\0')) {
    /* keyDBFile not found - put it in */
    PR_snprintf(buf, sizeof(buf), "keyDBFile: %s%c%s\n", 
                sroot, FILE_PATHSEP, key_file);
    lines[linecnt++] = PL_strdup(buf);
  }
  
  f = fopen(filename, "w");
  if (f==NULL) {
    fclose(f);
    servssl_error("Can not open adm.conf for writing");
  }
  
  for (i=0; i < linecnt; i++) {
    fprintf(f, "%s", lines[i]);
  }
  
  fclose(f);
}


/* NT SSL STUFF HERE -------------------------------------------------------*/

#ifdef XP_WIN32
static char password[512];
static char pinprompt[512];

//--------------------------------------------------------------------------//
// borrowed from \msvc20\mfc\src\dlgcore.cpp                                //
//--------------------------------------------------------------------------//
void static CenterDialog(HWND hwndParent, HWND hwndDialog)
{
   RECT DialogRect;
   RECT ParentRect;
        POINT Point;
   int nWidth;
   int nHeight;
 
        // Determine if the main window exists. This can be useful when
   // the application creates the dialog box before it creates the
   // main window. If it does exist, retrieve its size to center
   // the dialog box with respect to the main window.
   if(hwndParent != NULL)
        {
      GetClientRect(hwndParent, &ParentRect);
        }
   else
   {
           // if main window does not exist, center with respect to desktop
           hwndParent = GetDesktopWindow();
      GetWindowRect(hwndParent, &ParentRect);
   }
 
   // get the size of the dialog box
   GetWindowRect(hwndDialog, &DialogRect);
 
   // calculate height and width for MoveWindow()
   nWidth = DialogRect.right - DialogRect.left;
   nHeight = DialogRect.bottom - DialogRect.top;
 
   // find center point and convert to screen coordinates
   Point.x = (ParentRect.right - ParentRect.left) / 2;
   Point.y = (ParentRect.bottom - ParentRect.top) / 2;

        ClientToScreen(hwndParent, &Point);
 
   // calculate new X, Y starting point
   Point.x -= nWidth / 2;
   Point.y -= nHeight / 2;
 
   MoveWindow(hwndDialog, Point.x, Point.y, nWidth, nHeight, FALSE);
}



//--------------------------------------------------------------------------//
// length and wierdness checks are removed because they                     //
// do not belong in a password entry dialog.  This behavior OK-d            //
// by jsw@netscape.com on 01/26/96                                          //
//--------------------------------------------------------------------------//
BOOL CALLBACK PasswordDialogProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
   switch(message)
   {
   case WM_INITDIALOG:
     CenterDialog(NULL, hDlg);
     SendDlgItemMessage(hDlg, IDEDIT, EM_SETLIMITTEXT, sizeof(password), 0);
     SetDlgItemText(hDlg, IDTOKEN, pinprompt);
     EnableWindow(GetDlgItem(hDlg, IDOK), FALSE);
     return(FALSE);

   case WM_COMMAND:
     if(LOWORD(wParam) == IDEDIT)
     {
       if(HIWORD(wParam) == EN_CHANGE)
       {
         if(GetDlgItemText(hDlg, IDEDIT, password, sizeof(password)) > 0)
           EnableWindow(GetDlgItem(hDlg, IDOK), TRUE);
         else
           EnableWindow(GetDlgItem(hDlg, IDOK), FALSE);
       }
     }
     else
     if(LOWORD(wParam) == IDOK)
     {
       GetDlgItemText(hDlg, IDEDIT, password, sizeof(password));
       EndDialog(hDlg, IDOK);
       return (TRUE);
     }
     else
     if(LOWORD(wParam) == IDCANCEL)
     {
       /* User hit "Cancel" - terminate ns-admin.exe */
       memset(password, 0, sizeof(password));
       EndDialog(hDlg, IDCANCEL);
       exit(0);
     } else if (LOWORD(wParam) == IDHELP) {
       /* link will be broken if the file name is changed */
#define ADMSSLVIEWURL "start ..\\..\\bin\\base\\viewurl.exe "
#define ADMSSLFILE "file:///"
#define ADMSSLMANPATH "/manual/en/admin/help/5_secure.htm#SSL Token Password"
       char *my_netsite_root = getenv("NETSITE_ROOT");
       exec = PR_smprintf("%s%s%s%s", ADMSLVIEWURL, ADMSSLFILE, 
                                      my_netsite_root, ADMSSLMANPATH);
       if (NULL != exec) {
         system(exec);
         PR_smprintf_free(exec);
       }
     }
   }
   return (FALSE);
}

unsigned char *SEC_GetPassword(FILE *input, FILE *output, char *prompt,
                               PRBool (*ok)(unsigned char *))
{
   int result, retries = 0;

   PL_strncpyz(pinprompt, prompt, sizeof(pinprompt));
   result = DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_PASSWORD), 
                     HWND_DESKTOP, (int (__stdcall *)(void))PasswordDialogProc);
   retries = GetLastError();
   if (result == -1) {
      result = GetLastError();
      servssl_error("DialogBox failed (code %d)\n",
                    GetLastError());
      return NULL;
   }
   return (unsigned char*) PL_strdup((char*)password);
}
#endif   /* XP_WIN32 */


SECStatus MyAuthCertificateHook(void *arg, PRFileDesc *fd, PRBool checksig, PRBool isserver)
{
  return SECSuccess;
}


PR_IMPLEMENT(PRFileDesc*)
SSLSocket_init(PRFileDesc *req_socket)
{
    PRFileDesc *ssl_socket = NULL;
    char *serverRoot = getenv("NETSITE_ROOT");

    if (ADMSSL_InitSimple() != 0) {
       PR_Close(ssl_socket);
       return NULL;
    }

    if (SSL_OptionSetDefault(SSL_ENABLE_SSL2, PR_FALSE)) {
        return NULL;
    }

    ssl_socket = SSL_ImportFD(NULL, req_socket);

    SSL_AuthCertificateHook(ssl_socket, MyAuthCertificateHook, NULL);

    if (SSL_OptionSet(ssl_socket, SSL_SECURITY, PR_TRUE) < 0) {
        PR_Close(ssl_socket);
        return NULL;
    }

    return ssl_socket;
}
